@@ -19,10 +19,9 @@ import os_resource_classes as orc
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
 from oslo_utils import versionutils
-import six
 
-from nova.db.sqlalchemy import api as db
-from nova.db.sqlalchemy import api_models
+from nova.db.api import api as api_db_api
+from nova.db.api import models as api_models
 from nova import exception
 from nova import objects
 from nova.objects import base
@@ -35,7 +34,8 @@ REQUEST_SPEC_OPTIONAL_ATTRS = ['requested_destination',
                                'security_groups',
                                'network_metadata',
                                'requested_resources',
-                               'request_level_params']
+                               'request_level_params',
+                               'requested_networks']
 
 
 @base.NovaObjectRegistry.register
@@ -54,7 +54,8 @@ class RequestSpec(base.NovaObject):
     # Version 1.11: Added is_bfv
     # Version 1.12: Added requested_resources
     # Version 1.13: Added request_level_params
-    VERSION = '1.13'
+    # Version 1.14: Added requested_networks
+    VERSION = '1.14'
 
     fields = {
         'id': fields.IntegerField(),
@@ -109,11 +110,16 @@ class RequestSpec(base.NovaObject):
                                                          default=None),
         # NOTE(efried): This field won't be persisted.
         'request_level_params': fields.ObjectField('RequestLevelParams'),
+        # NOTE(sbauza); This field won't be persisted. For move operations, we
+        # reevaluate it using the network-related instance info_cache.
+        'requested_networks': fields.ObjectField('NetworkRequestList')
     }
 
     def obj_make_compatible(self, primitive, target_version):
         super(RequestSpec, self).obj_make_compatible(primitive, target_version)
         target_version = versionutils.convert_version_to_tuple(target_version)
+        if target_version < (1, 14) and 'requested_networks' in primitive:
+            del primitive['requested_networks']
         if target_version < (1, 13) and 'request_level_params' in primitive:
             del primitive['request_level_params']
         if target_version < (1, 12):
@@ -153,6 +159,10 @@ class RequestSpec(base.NovaObject):
             self.request_level_params = RequestLevelParams()
             return
 
+        if attrname == 'requested_networks':
+            self.requested_networks = objects.NetworkRequestList(objects=[])
+            return
+
         # NOTE(sbauza): In case the primitive was not providing that field
         # because of a previous RequestSpec version, we want to default
         # that field in order to have the same behaviour.
@@ -190,6 +200,12 @@ class RequestSpec(base.NovaObject):
         # respective obj_load_attr methods.
         return self.request_level_params.root_forbidden
 
+    @property
+    def same_subtree(self):
+        # self.request_level_params and .same_subtree lazy-default via their
+        # respective obj_load_attr methods.
+        return self.request_level_params.same_subtree
+
     def _image_meta_from_image(self, image):
         if isinstance(image, objects.ImageMeta):
             self.image = image
@@ -232,14 +248,14 @@ class RequestSpec(base.NovaObject):
 
     def _from_instance_pci_requests(self, pci_requests):
         if isinstance(pci_requests, dict):
-            pci_req_cls = objects.InstancePCIRequests
-            self.pci_requests = pci_req_cls.from_request_spec_instance_props(
-                pci_requests)
+            self.pci_requests = objects.InstancePCIRequests.obj_from_primitive(
+                pci_requests,
+            )
         else:
             self.pci_requests = pci_requests
 
     def _from_instance_numa_topology(self, numa_topology):
-        if isinstance(numa_topology, six.string_types):
+        if isinstance(numa_topology, str):
             numa_topology = objects.InstanceNUMATopology.obj_from_primitive(
                 jsonutils.loads(numa_topology))
 
@@ -458,10 +474,12 @@ class RequestSpec(base.NovaObject):
         return filt_props
 
     @classmethod
-    def from_components(cls, context, instance_uuid, image, flavor,
-            numa_topology, pci_requests, filter_properties, instance_group,
-            availability_zone, security_groups=None, project_id=None,
-            user_id=None, port_resource_requests=None):
+    def from_components(
+        cls, context, instance_uuid, image, flavor,
+        numa_topology, pci_requests, filter_properties, instance_group,
+        availability_zone, security_groups=None, project_id=None,
+        user_id=None, port_resource_requests=None, request_level_params=None
+    ):
         """Returns a new RequestSpec object hydrated by various components.
 
         This helper is useful in creating the RequestSpec from the various
@@ -487,6 +505,7 @@ class RequestSpec(base.NovaObject):
         :param port_resource_requests: a list of RequestGroup objects
                                        representing the resource needs of the
                                        neutron ports
+        :param request_level_params: a RequestLevelParams object
         """
         spec_obj = cls(context)
         spec_obj.num_instances = 1
@@ -520,9 +539,11 @@ class RequestSpec(base.NovaObject):
         if port_resource_requests:
             spec_obj.requested_resources.extend(port_resource_requests)
 
-        # NOTE(efried): We don't need to handle request_level_params here yet
-        #  because they're set dynamically by the scheduler. That could change
-        #  in the future.
+        # NOTE(gibi): later the scheduler adds more request level params but
+        # never overrides existing ones so we can initialize them here.
+        if request_level_params is None:
+            request_level_params = objects.RequestLevelParams()
+        spec_obj.request_level_params = request_level_params
 
         # NOTE(sbauza): Default the other fields that are not part of the
         # original contract
@@ -535,12 +556,16 @@ class RequestSpec(base.NovaObject):
         if 'user_id' not in self or self.user_id is None:
             self.user_id = instance.user_id
 
-    def ensure_network_metadata(self, instance):
+    def ensure_network_information(self, instance):
         if not (instance.info_cache and instance.info_cache.network_info):
+            # NOTE(sbauza): On create, the network_info field is null but we
+            # directly set the RequestSpec nested network_requests field, so we
+            # are fine returning here.
             return
 
         physnets = set([])
         tunneled = True
+        network_requests = []
 
         # physical_network and tunneled might not be in the cache for old
         # instances that haven't had their info_cache healed yet
@@ -552,19 +577,35 @@ class RequestSpec(base.NovaObject):
             tunneled |= vif.get('network', {}).get('meta', {}).get(
                 'tunneled', False)
 
+            # We also want to recreate the original NetworkRequests
+            # TODO(sbauza): We miss tag and pci_request_id information that is
+            # not stored in the VIF model to fully provide all fields
+            # FIXME(sbauza): We can't also guess whether the user provided us
+            # a specific IP address to use for create, and which one.
+            nr_args = {
+                'network_id': vif['network']['id'],
+                'port_id': vif['id'],
+            }
+            network_request = objects.NetworkRequest(**nr_args)
+            network_requests.append(network_request)
         self.network_metadata = objects.NetworkMetadata(
             physnets=physnets, tunneled=tunneled)
+        self.requested_networks = objects.NetworkRequestList(
+            objects=network_requests)
 
     @staticmethod
     def _from_db_object(context, spec, db_spec):
         spec_obj = spec.obj_from_primitive(jsonutils.loads(db_spec['spec']))
+        data_migrated = False
+
         for key in spec.fields:
             # Load these from the db model not the serialized object within,
             # though they should match.
             if key in ['id', 'instance_uuid']:
                 setattr(spec, key, db_spec[key])
             elif key in ('requested_destination', 'requested_resources',
-                         'network_metadata', 'request_level_params'):
+                         'network_metadata', 'request_level_params',
+                         'requested_networks'):
                 # Do not override what we already have in the object as this
                 # field is not persisted. If save() is called after
                 # one of these fields is populated, it will reset the field to
@@ -577,6 +618,13 @@ class RequestSpec(base.NovaObject):
                 # fields. If they are not set, set None.
                 if not spec.obj_attr_is_set(key):
                     setattr(spec, key, None)
+            elif key == "numa_topology":
+                if key in spec_obj:
+                    spec.numa_topology = spec_obj.numa_topology
+                    if spec.numa_topology:
+                        data_migrated = objects.InstanceNUMATopology.\
+                            _migrate_legacy_dedicated_instance_cpuset(
+                                spec.numa_topology)
             elif key in spec_obj:
                 setattr(spec, key, getattr(spec_obj, key))
         spec._context = context
@@ -597,12 +645,16 @@ class RequestSpec(base.NovaObject):
             except exception.InstanceGroupNotFound:
                 # NOTE(danms): Instance group may have been deleted
                 spec.instance_group = None
+                spec.scheduler_hints.pop('group', None)
+
+        if data_migrated:
+            spec.save()
 
         spec.obj_reset_changes()
         return spec
 
     @staticmethod
-    @db.api_context_manager.reader
+    @api_db_api.context_manager.reader
     def _get_by_instance_uuid_from_db(context, instance_uuid):
         db_spec = context.session.query(api_models.RequestSpec).filter_by(
             instance_uuid=instance_uuid).first()
@@ -617,7 +669,7 @@ class RequestSpec(base.NovaObject):
         return cls._from_db_object(context, cls(), db_spec)
 
     @staticmethod
-    @db.api_context_manager.writer
+    @api_db_api.context_manager.writer
     def _create_in_db(context, updates):
         db_spec = api_models.RequestSpec()
         db_spec.update(updates)
@@ -646,14 +698,21 @@ class RequestSpec(base.NovaObject):
                 spec.instance_group.hosts = None
             # NOTE(mriedem): Don't persist these since they are per-request
             for excluded in ('retry', 'requested_destination',
-                             'requested_resources', 'ignore_hosts',
-                             'request_level_params'):
+                             'requested_resources', 'ignore_hosts'):
                 if excluded in spec and getattr(spec, excluded):
                     setattr(spec, excluded, None)
             # NOTE(stephenfin): Don't persist network metadata since we have
             # no need for it after scheduling
             if 'network_metadata' in spec and spec.network_metadata:
                 del spec.network_metadata
+            # NOTE(sbauza): Don't persist requested_networks since we have
+            # no need for it after scheduling
+            if 'requested_networks' in spec and spec.requested_networks:
+                del spec.requested_networks
+            # NOTE(gibi): Don't persist requested_networks since we have
+            # no need for it after scheduling
+            if 'request_level_params' in spec and spec.request_level_params:
+                del spec.request_level_params
 
             db_updates = {'spec': jsonutils.dumps(spec.obj_to_primitive())}
             if 'instance_uuid' in updates:
@@ -674,7 +733,7 @@ class RequestSpec(base.NovaObject):
         self._from_db_object(self._context, self, db_spec)
 
     @staticmethod
-    @db.api_context_manager.writer
+    @api_db_api.context_manager.writer
     def _save_in_db(context, instance_uuid, updates):
         # FIXME(sbauza): Provide a classmethod when oslo.db bug #1520195 is
         # fixed and released
@@ -694,7 +753,7 @@ class RequestSpec(base.NovaObject):
             self.obj_reset_changes()
 
     @staticmethod
-    @db.api_context_manager.writer
+    @api_db_api.context_manager.writer
     def _destroy_in_db(context, instance_uuid):
         result = context.session.query(api_models.RequestSpec).filter_by(
             instance_uuid=instance_uuid).delete()
@@ -706,7 +765,7 @@ class RequestSpec(base.NovaObject):
         self._destroy_in_db(self._context, self.instance_uuid)
 
     @staticmethod
-    @db.api_context_manager.writer
+    @api_db_api.context_manager.writer
     def _destroy_bulk_in_db(context, instance_uuids):
         return context.session.query(api_models.RequestSpec).filter(
                 api_models.RequestSpec.instance_uuid.in_(instance_uuids)).\
@@ -1144,6 +1203,54 @@ class RequestGroup(base.NovaEphemeralObject):
         obj.obj_set_defaults()
         return obj
 
+    @classmethod
+    def from_extended_port_request(cls, context, port_resource_request):
+        """Create the group objects from the resource request of a neutron port
+
+        :param context: the request context
+        :param port_resource_request: the resource_request attribute of the
+                                      neutron port
+        For example:
+
+            port_resource_request = {
+                "request_groups":
+                [
+                    {
+                        "id": <uuid>
+                        "required": [CUSTOM_PHYSNET_2,
+                                     CUSTOM_VNIC_TYPE_NORMAL],
+                        "resources":{
+                            NET_PACKET_RATE_KILOPACKET_PER_SEC: 1000
+                        }
+                    },
+                    {
+                        "id": <uuid>
+                        "required": [CUSTOM_PHYSNET_2,
+                                     CUSTOM_VNIC_TYPE_NORMAL],
+                        "resources": {
+                            "NET_BW_IGR_KILOBIT_PER_SEC": 1000,
+                            "NET_BW_EGR_KILOBIT_PER_SEC": 1000,
+                        },
+                    }
+                ]
+            }
+        """
+        group_objs = []
+        for group in port_resource_request.get("request_groups", []):
+            # NOTE(gibi): Placement rejects allocation candidates where a
+            # request group has traits but no resources specified. This is why
+            # resources are handled as mandatory below but not traits.
+            obj = cls(
+                context=context,
+                use_same_provider=True,
+                resources=group['resources'],
+                required_traits=set(group.get('required', [])),
+                requester_id=group['id'])
+            obj.obj_set_defaults()
+            group_objs.append(obj)
+
+        return group_objs
+
     def obj_make_compatible(self, primitive, target_version):
         super(RequestGroup, self).obj_make_compatible(
             primitive, target_version)
@@ -1205,6 +1312,51 @@ class RequestGroup(base.NovaEphemeralObject):
             if self.resources[rclass] == 0:
                 self.resources.pop(rclass)
 
+    def to_queryparams(self):
+        """Convert the RequestGroup to placement allocation candidates query
+        parameters.
+        """
+
+        # NOTE(efried): The sorting herein is not necessary for the API; it is
+        # to make testing easier and logging/debugging predictable.
+        res = self.resources
+        required_traits = self.required_traits
+        forbidden_traits = self.forbidden_traits
+        aggregates = self.aggregates
+        in_tree = self.in_tree
+        forbidden_aggregates = self.forbidden_aggregates
+        suffix = self.requester_id or ''
+
+        resource_query = ",".join(
+            sorted("%s:%s" % (rc, amount)
+                   for (rc, amount) in res.items()))
+        qs_params = [('resources%s' % suffix, resource_query)]
+
+        # Assemble required and forbidden traits, allowing for either/both
+        # to be empty.
+        required_val = ','.join(
+            sorted(required_traits) +
+            ['!%s' % ft for ft in sorted(forbidden_traits)])
+        if required_val:
+            qs_params.append(('required%s' % suffix, required_val))
+        if aggregates:
+            aggs = []
+            # member_of$S is a list of lists.  We need a tuple of
+            # ('member_of$S', 'in:uuid,uuid,...') for each inner list.
+            for agglist in aggregates:
+                aggs.append(('member_of%s' % suffix,
+                             'in:' + ','.join(sorted(agglist))))
+            qs_params.extend(sorted(aggs))
+        if in_tree:
+            qs_params.append(('in_tree%s' % suffix, in_tree))
+        if forbidden_aggregates:
+            # member_of$S is a list of aggregate uuids. We need a
+            # tuple of ('member_of$S, '!in:uuid,uuid,...').
+            forbidden_aggs = '!in:' + ','.join(
+                sorted(forbidden_aggregates))
+            qs_params.append(('member_of%s' % suffix, forbidden_aggs))
+        return qs_params
+
 
 @base.NovaObjectRegistry.register
 class RequestLevelParams(base.NovaObject):
@@ -1213,16 +1365,55 @@ class RequestLevelParams(base.NovaObject):
     RequestGroup).
     """
     # Version 1.0: Initial version
-    VERSION = '1.0'
+    # Version 1.1: Add same_subtree field
+    VERSION = '1.1'
 
     fields = {
         # Traits required on the root provider
         'root_required': fields.SetOfStringsField(default=set()),
         # Traits forbidden on the root provider
         'root_forbidden': fields.SetOfStringsField(default=set()),
+        # Lists of request group suffixes that needs to be allocated from the
+        # same provider subtree
+        'same_subtree': fields.ListOfListsOfStringsField(default=list()),
         # NOTE(efried): group_policy would be appropriate to include here, once
         # we have a use case for it.
     }
 
     def obj_load_attr(self, attrname):
         self.obj_set_defaults(attrname)
+
+    def obj_make_compatible(self, primitive, target_version):
+        super().obj_make_compatible(primitive, target_version)
+
+        target_version = versionutils.convert_version_to_tuple(target_version)
+        if target_version < (1, 1):
+            if 'same_subtree' in primitive:
+                del primitive['same_subtree']
+
+    @classmethod
+    def from_port_request(cls, port_resource_request):
+        """Extracts request level global parameters from the resource_request
+        of a neutron port.
+
+        Neutron only uses same_subtree at the moment.
+        """
+        same_subtree = port_resource_request.get('same_subtree')
+        if same_subtree:
+            # NOTE(gibi): A single port only has a single list of groups
+            # requesting same subtree, but the RequestLevelParams maintains a
+            # list of such subtree requests
+            return cls(same_subtree=[same_subtree])
+
+        return cls()
+
+    def extend_with(self, other_req_lvl_params):
+        """Extends the existing object with parameter values from another
+        RequestLevelParams object. That new set of requirements are connected
+        with the existing ones with a logical AND operation.
+        """
+        self.root_required = self.root_required.union(
+            other_req_lvl_params.root_required)
+        self.root_forbidden = self.root_forbidden.union(
+            other_req_lvl_params.root_forbidden)
+        self.same_subtree.extend(other_req_lvl_params.same_subtree)
